/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.Constants.Interceptor.Phase;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.performance.PerformanceNotification;
import org.ow2.petals.component.framework.su.ServiceUnitManager;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * The abstract methods for a CDK listener. Both external and JBI listner extend
 * it.
 * 
 * @author chamerling - EBM WebSourcing
 * @author ofabre - EBM WebSourcing
 * @author alouis - EBM WebSourcing
 * 
 */
public abstract class AbstractListener {

    /**
     * The JBI component itself
     */
    protected AbstractComponent component;

    /**
     * The consumes
     */
    protected Consumes consumes;

    /**
     * The provides
     */
    protected Provides provides;

    /**
     * The Consumes or Provides extensions
     */
    private ConfigurationExtensions extensions;

    /**
     * The message interceptor processor; Each listener have its own interceptor
     * processor in order to get better performances.
     */
    protected MessageInterceptorProcessor messageInterceptorProcessor;

    /**
     * The messageExchange factory used to create MessageExchanges
     */
    protected MessageExchangeFactory factory;

    /**
     * The current Transaction
     */
    private Transaction currentTransaction;

    /**
     * Creates a new instance of {@link AbstractListener}
     * 
     */
    public AbstractListener() {
        super();
    }

    /**
     * Initialize the listener.
     * 
     * @param component
     */
    public final void init(AbstractComponent component) {
        this.component = component;
        this.messageInterceptorProcessor = new MessageInterceptorProcessor(this.component, this
                .getLogger());
        this.factory = this.getComponent().getChannel().createExchangeFactory();
    }

    /**
     * Create an exchange implementation matching the given pattern.
     * 
     * @param mep
     *      a specific MEP pattern
     * @return an exchange implementation matching the given pattern
     * @throws MessagingException
     *      The message exchange can't be created or no pattern specified in
     *      extensions
     * @throws PEtALSCDKException
     *      The current Transaction has failed to return its status
     */
    public Exchange createExchange(MEPConstants mep) throws MessagingException, PEtALSCDKException {
        MessageExchange exchange = this.factory.createExchange(mep.value());

        // if there is a current transaction,
        try {
            if (this.currentTransaction != null
                    && this.currentTransaction.getStatus() == Status.STATUS_ACTIVE) {
                exchange.setProperty(MessageExchange.JTA_TRANSACTION_PROPERTY_NAME,
                        this.currentTransaction);
            }
        } catch (SystemException e) {
            throw new PEtALSCDKException("Failed to get current transaction", e);
        }

        return new ExchangeImpl(exchange);
    }

    /**
     * Call accept(MessageExchange exchange, long timeout) with a timeout of "0"
     * that means no timeout
     * 
     * @param exchange
     *      the exchange that has been sent in an asynchronous way
     * @return the response {@link MessageExchange} to an asynchronous service
     *  call
     * @throws InterruptedException
     * @throws PEtALSCDKException
     *      if this method is called despite the fact that response
     *      receptiongetDeliveryChannel has been acknowledged
     */
    public Exchange accept(Exchange exchange) throws InterruptedException, PEtALSCDKException {
        return this.accept(exchange, 0);
    }

    /**
     * This method blocks until response to the given {@link MessageExchange}
     * reaches the component. A timeout must be specified. Waiting stops when
     * timeout occured.
     * 
     * @param exchange
     *      the exchange that has been sent in an asynchronous way
     * @param timeout
     *      waiting message return until timeout occured. Timeout in
     *      milliseconds. "0" means no timeout
     * @return the response {@link MessageExchange} to an asynchronous service
     *  call. Can be null if timeout occured.
     * @throws InterruptedException
     * @throws PEtALSCDKException
     *      if this method is called despite the fact that response reception
     *      has been acknowledged
     */
    public Exchange accept(Exchange exchange, long timeout) throws InterruptedException,
            PEtALSCDKException {
        // Try to find the response in the synchronizer (response arrives before
        // the call to accept)
        this.component.getLogger().finest(
                "Wait for a response to exchange of id: " + exchange.getExchangeId());
        MessageExchangeSynchronizer exchangeSynchronizer = this.component.getAsyncMessageHandler()
                .getAsynchronousMessage(exchange.getExchangeId());
        Exchange responseExchange = null;
        if (exchangeSynchronizer != null) {
            responseExchange = exchangeSynchronizer.getExchange();
            if (responseExchange == null) {
                // No response arrived, so block until response arrival
                responseExchange = exchangeSynchronizer.accept(timeout);
            }
            exchangeSynchronizer.setAccepted(true);
        } else {
            throw new PEtALSCDKException("You are trying to accept a response for"
                    + " an exchange not synchronized");
        }
        this.component.getLogger().finest(
                "Asynchronous response arrived for exchange with id : " + exchange.getExchangeId());
        return responseExchange;
    }

    /**
     * This method must be called to enable the CDK to automatically send an
     * acknowledgment to the provider of the service, at the end of a CDK
     * managed asynchronous exchange. It automatically set exchange status to
     * DONE if exchange pattern is RobustInOnly or InOut or if exchange pattern
     * is InOptOut and no fault is set.
     * 
     * @param exchange
     *      the message exchange originating the response synchronization
     * @throws MessagingException
     *      if an error occurs during acknowledgment
     * @throws PEtALSCDKException
     *      if you try close exchange before accepting response or for an
     *      unsynchronized exchange
     */
    public void closeExchange(Exchange exchange) throws MessagingException, PEtALSCDKException {
        MessageExchangeSynchronizer exchangeSynchronizer = this.component.getAsyncMessageHandler()
                .getAsynchronousMessage(exchange.getExchangeId());
        if (exchangeSynchronizer != null) {
            if (exchangeSynchronizer.isAccepted() && !exchangeSynchronizer.isTimeout()) {
                Exchange returnedMessage = exchangeSynchronizer.getExchange();
                this.processAck(returnedMessage);
                exchangeSynchronizer.ackResponseProcessed();
                this.component.getChannel().send(
                        ((ExchangeImpl) returnedMessage).getMessageExchange());
                this.component.getLogger().finest(
                        "Close exchange with id : " + exchange.getExchangeId());
            } else {
                throw new PEtALSCDKException("You are trying to close an exchange despite "
                        + "the fact that you have not received response "
                        + "(call accept() method before " + "or specify a longest timeout)");
            }
        } else {
            throw new PEtALSCDKException("You are trying to close an exchange not synchronized");
        }
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/> Process MessageExchange interceptors before
     * sending message. <br/> No synchronization mecanism activated
     * 
     * @param exchange
     *      message exchange to send; must be non-null
     * @throws MessagingException
     *      if the address has no endpoint associated to it, or if an error
     *      occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *  createConsumeExchange()
     */
    public void send(final Exchange exchange) throws MessagingException {

        try {
            this.interceptSend(exchange);
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        }

        this.component.getLogger().finest(
                "Send an asynchronous exchange with id : " + exchange.getExchangeId());
        this.getComponent().getChannel().send(((ExchangeImpl) exchange).getMessageExchange());
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/> Process MessageExchange interceptors before
     * sending message. <br/> Add a synchronization mecanism that enable usage
     * of a blocking accept(messageId) method that will be unlock when the
     * response message with the same id reaches the component
     * 
     * @param exchange
     *      message exchange to send; must be non-null
     * @throws MessagingException
     *      if the address has no endpoint associated to it, or if an error
     *      occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *  createConsumeExchange()
     */
    public void sendAsync(final Exchange exchange) throws MessagingException {
        // Register a synchroniser
        MessageExchangeSynchronizer exchangeSynchronizer = new MessageExchangeSynchronizer();
        this.component.getAsyncMessageHandler().addAsynchronousMessage(exchange.getExchangeId(),
                exchangeSynchronizer);

        this.send(exchange);
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * synchronous way.<br/> Process MessageExchange interceptors before sending
     * message. <br/> Add performance notification
     * 
     * @param exchange
     *      message exchange to send; must be non-null
     * @param timeoutMS
     *      timeout period, in milliseconds (0 means no timeout); must be
     *      greater than or equal to zero
     * @return true if the message has been sent (if a timeout occurs, return
     *  false)
     * @throws MessagingException
     *      if the address has no endpoint associated to it, or if an error
     *      occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *  createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange, long timeoutMS) throws MessagingException {
        boolean bRet = false;

        try {
            this.interceptSend(exchange);

            // Notification creation
            PerformanceNotification performanceNotification = this.component
                    .createPerformanceNotificationSendingToNMR(this
                            .getPerformanceNotificationUserData(exchange), exchange);

            this.component.getLogger().finest(
                    "Send a synchronous exchange with id : " + exchange.getExchangeId());
            bRet = this.getComponent().getChannel().sendSync(
                    ((ExchangeImpl) exchange).getMessageExchange(), timeoutMS);

            // Notification sending
            this.component.sendPerformanceNotification(performanceNotification);

            this.interceptSendResponse(exchange);
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        } catch (JBIException e) {
            throw new MessagingException(e);
        }

        return bRet;
    }

    /**
     * Call sendSync(MessageExchage exchange, long timeoutMS) with a timeout
     * value found in provides or consumes extensions or "0" if not found.
     * 
     * @return true if the message has been sent (if a timeout occurs, return
     *         false)
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange) throws MessagingException {
        long timeoutMS = 0;

        // set the timeout if it is configured
        if (this.consumes != null) {
            if (this.consumes.getTimeout() != null) {
                timeoutMS = Long.valueOf(this.consumes.getTimeout());
            }
        } else if (this.provides != null) {
            if (this.provides.getTimeout() != null) {
                timeoutMS = Long.valueOf(this.provides.getTimeout());
            }
        }

        return this.sendSync(exchange, timeoutMS);
    }

    /**
     * Create a message exchange with all the information set in the CONSUME
     * block of the related serviceunit: <br>
     * - service information (itf, srv, ep) <br>
     * - MEP <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *      The message exchange can't be created or no pattern specified in
     *      extensions
     * @throws PEtALSCDKException
     *      The current Transaction has failed to return its status
     * @throws PEtALSCDKException
     */
    public Exchange createConsumeExchange() throws MessagingException, PEtALSCDKException {
        return createConsumeExchange(consumes);
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME <br>
     * - address (interface, service, endpoint) <br>
     * - MEP <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *      The message exchange can't be created or no pattern specified in
     *      extensions
     * @throws PEtALSCDKException
     *      The current Transaction has failed to return its status
     */
    public Exchange createConsumeExchange(Consumes consumes) throws MessagingException,
            PEtALSCDKException {

        MEPConstants mep = this.retrievePattern(consumes);
        Exchange exchange = this.createConsumeExchange(consumes, mep);
        return exchange;
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME, with the given exchange pattern <br>
     * - address (interface, service, endpoint) <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *      The message exchange can't be created or no pattern specified in
     *      extensions
     * @throws PEtALSCDKException
     *      The current Transaction has failed to return its status
     */
    public Exchange createConsumeExchange(Consumes consumes, MEPConstants mep)
            throws MessagingException, PEtALSCDKException {

        Exchange exchange = this.createExchange(mep);
        this.setOperationFromConsumes(exchange, consumes);
        this.setMessageExchangePropertiesFromConsumes(exchange, consumes);
        this.setExchangeDestinationForConsumes(exchange, consumes);

        return exchange;
    }

    /**
     * Get the service extensions.
     * 
     * @return Extensions, can be null if no extensions found
     */
    public ConfigurationExtensions getExtensions() {
        return this.extensions;
    }

    /**
     * The direct usage of the Component context is not recommended in the CDK.
     * 
     * @return
     */
    @Deprecated
    public ComponentContext getComponentContext() {
        return this.component.getContext();
    }

    /**
     * 
     * @return
     */
    @Deprecated
    public DeliveryChannel getDeliveryChannel() {
        return this.component.getChannel();
    }

    /**
     * 
     * @return
     */
    public Logger getLogger() {
        return this.component.getLogger();
    }

    /**
     * 
     * @return
     */
    public AbstractComponent getComponent() {
        return this.component;
    }

    /**
     * @return the consumes
     */
    public Consumes getConsumes() {
        return this.consumes;
    }

    public String getInstallRoot() {
        ServiceUnitManager suManager = (ServiceUnitManager) this.getComponent()
                .getServiceUnitManager();
        String installRoot = null;
        if (this.consumes != null) {
            installRoot = suManager.getSUDataHandlerForConsumes(this.consumes).getInstallRoot();
        } else {
            installRoot = suManager.getSUDataHandlerForProvides(this.provides).getInstallRoot();
        }
        return installRoot;
    }

    /**
     * Return the current Transaction.
     * 
     * @return The current Transaction or {@code null} if no Transaction Manager
     *  is implemented by the JBI container
     * 
     * @throws SystemException
     * @throws NotSupportedException
     */
    public Transaction getCurrentTransaction() throws PEtALSCDKException {
        if (this.currentTransaction == null) {
            TransactionManager transactionManager = (TransactionManager) this.component
                    .getContext().getTransactionManager();
            if (transactionManager != null) {
                try {
                    transactionManager.begin();
                    this.currentTransaction = transactionManager.getTransaction();
                } catch (NotSupportedException e) {
                    new PEtALSCDKException("Failed to create a Transaction", e);
                } catch (SystemException e) {
                    new PEtALSCDKException("Failed to create a Transaction", e);
                }

            }
        }
        return this.currentTransaction;
    }

    /**
     * @return the messageInterceptorProcessor
     */
    public MessageInterceptorProcessor getMessageInterceptorProcessor() {
        return this.messageInterceptorProcessor;
    }

    /**
     * @param consumes
     *      the consumes to set
     */
    protected void setConsumes(Consumes consumes) {
        this.consumes = consumes;
    }

    /**
     * @param consumes
     *      the consumes to set
     */
    protected void setExtensions(ConfigurationExtensions extensions) {
        this.extensions = extensions;
    }

    /**
     * This method should be overriden if you wish to set custom user data to
     * your performance notification. By default, the user data is set to the
     * mesasge exchange endpoint name if exists, else with the service name,
     * else with the interface name.
     * 
     * @param exchange
     * @return
     */
    protected String getPerformanceNotificationUserData(final Exchange exchange) {
        String sRet = null;
        if (exchange.getEndpoint() != null) {
            sRet = exchange.getEndpoint().toString();
        }

        if (sRet == null) {
            if (exchange.getService() != null) {
                sRet = exchange.getService().toString();
            }
        }

        if (sRet == null) {
            if (exchange.getInterfaceName() != null) {
                sRet = exchange.getInterfaceName().toString();
            }
        }

        return sRet;
    }

    /**
     * Set operation to call if exists in extensions. If no extension can be
     * found or if operation value is "" or missing, operation isn't set.
     * 
     * @param exchange
     *      a {@link MessageExchange}, must not be null
     * @param extensions
     */
    protected void setOperationFromConsumes(final Exchange exchange, Consumes consumes) {

        if (!StringHelper.isNullOrEmpty(consumes.getOperation())) {
            QName operation = QName.valueOf(consumes.getOperation());
            exchange.setOperation(operation);
        }
    }

    /**
     * Set message exchange properties from extensions. All extensions key
     * starting with "org.ow2.petals" are automatically added to message
     * exchange properties.
     * 
     * @param exchange
     *      a {@link Exchange}, must not be null
     * @param extensions
     * @throws PEtALSCDKException
     */
    protected void setMessageExchangePropertiesFromConsumes(final Exchange exchange,
            final Consumes consumes) {

        if (consumes.getExchangeProperties() != null) {
            ConfigurationExtensions configurationExtensions = null;
            if (this.getComponent().getComponentConfiguration().getPropertiesFile() != null) {
                try {
                    configurationExtensions = new ConfigurationExtensions(consumes
                            .getExchangeProperties().getExchangeProperty(), this.getComponent()
                            .getComponentConfiguration().getPropertiesFile().getValue());
                } catch (PEtALSCDKException e) {
                    this.getLogger().warning(
                            "Failed to initialize Message Exchanges properties against component properties: "
                                    + e.getMessage());
                }
            } else {
                configurationExtensions = new ConfigurationExtensions(consumes
                        .getExchangeProperties().getExchangeProperty());
            }
            if (configurationExtensions != null) {
                for (String name : configurationExtensions.keySet()) {
                    exchange.setProperty(name, configurationExtensions.get(name));
                }
            }
        }
    }

    protected void setCurrentTransaction(Transaction transaction) {
        this.currentTransaction = transaction;
    }

    /**
     * Retrieve the MEP from the extensions.
     * 
     * @param extensions
     * @return mep specified in extensions
     * @throws MessagingException
     *      MEP pattern specified in consumes extension isn't valid or if no MEP
     *      pattern specified in extensions.
     */
    protected final MEPConstants retrievePattern(Consumes consumes) throws MessagingException {
        MEPConstants mep = null;

        if (consumes.getMep() != null) {
            if (MEPType.IN_ONLY.equals(consumes.getMep())) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else if (MEPType.ROBUST_IN_ONLY.equals(consumes.getMep())) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (MEPType.IN_OUT.equals(consumes.getMep())) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (MEPType.IN_OPTIONAL_OUT.equals(consumes.getMep())) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            }
        } else {
            throw new MessagingException("No MEP pattern specified in extensions");
        }

        return mep;
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.init(). This method is called after setting of all Listener
     * attributes
     * 
     */
    protected void init() {
    }

    private void processAck(Exchange exchange) throws MessagingException {
        if (exchange.isRobustInOnlyPattern() || exchange.isInOutPattern()) {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        } else if (exchange.isInOptionalOutPattern()) {
            if (exchange.getFault() == null && exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * Fill the exchange with the service information (itf,src, and ep) from the
     * consume extension
     * 
     * @param exchange
     */
    private void setExchangeDestinationForConsumes(final Exchange exchange, final Consumes consumes) {
        
        // Retrieve possible target endpoint
        ServiceEndpoint ep = null;
        if (consumes.getEndpointName() != null && consumes.getServiceName() != null) {
            ep = this.getComponent().getContext().getEndpoint(consumes.getServiceName(),
                    consumes.getEndpointName());
        }
        
        // set exchange information
        exchange.setInterfaceName(consumes.getInterfaceName());
        exchange.setService(consumes.getServiceName());
        exchange.setEndpoint(ep);

        if (consumes.getLinkType() != null) {
            exchange.setProperty(Exchange.PROPERTY_PROVIDER_LINKTYPE, consumes.getLinkType().value());
        }
    }

    /**
     * Call the Interceptors before sending the request exchange
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptSend(final Exchange exchange) throws PEtALSCDKException {
        if (this.consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, this.consumes, Phase.SEND);
        }
    }

    /**
     * Call the Interceptors on exchange response
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptSendResponse(final Exchange exchange) throws PEtALSCDKException {
        if (this.consumes != null) {
            this.messageInterceptorProcessor
                    .intercept(exchange, this.consumes, Phase.SEND_RESPONSE);
        }
    }
}
