/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.su;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptors;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settableboolean;
import org.ow2.petals.component.framework.util.IDGenerator;
import org.ow2.petals.component.framework.util.ManagementMessageUtil;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.WSDLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import static org.ow2.petals.component.framework.Constants.Component.AUTOGENERATED_ENDPOINT_NAME;
import static org.ow2.petals.component.framework.Constants.Component.JBI_XML;
import static org.ow2.petals.component.framework.Constants.Component.META_INF;

/**
 * The common part of the ServiceUnitManager. It handles service units JBI
 * descriptor during their deployment, life-cycle and undeployment.
 * 
 * @author ofabre - EBM Websourcing
 * @author alouis - EBM Websourcing
 * 
 */
public abstract class ServiceUnitManager implements javax.jbi.component.ServiceUnitManager {

    /**
     * The component's Logger
     */
    protected Logger logger;

    /**
     * The service unit data handlers. Each service unit is linked to a
     * datahandler. Key is the service unit name.
     */
    private final ConcurrentMap<String, ServiceUnitDataHandler> serviceUnitDataHandlers;

    protected final AbstractComponent component;

    /**
     * Creates a new instance of {@link ServiceUnitManager}
     * 
     * @param component
     */
    public ServiceUnitManager(final AbstractComponent component) {
        super();
        this.component = component;
        this.logger = component.getLogger();
        this.serviceUnitDataHandlers = new ConcurrentHashMap<String, ServiceUnitDataHandler>();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#init(java.lang.String,
     * java.lang.String)
     */
    public void init(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.FINE, "init  serviceUnitName  " + serviceUnitName
                + "  serviceUnitRootPath " + suRootPath);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#deploy(java.lang.String,
     *  java.lang.String)
     */
    public String deploy(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        if (this.component.getContext() == null) {
            // the component has not been initialized for the first time...
            throw new DeploymentException(
                    "Can not deploy a service unit on a component which has not been initialized");
        }

        this.logger.log(Level.FINE, "deploy  serviceUnitName  " + serviceUnitName
                + "serviceUnitRootPath" + suRootPath);

        final File jbiXmlFile = new File(suRootPath + File.separator + META_INF + File.separator
                + JBI_XML);
        Jbi jbiDescriptor;
        try {
            final InputStream jbiInputStream = new FileInputStream(jbiXmlFile);
            try {
                jbiDescriptor = JBIDescriptorBuilder.buildJavaJBIDescriptor(jbiInputStream);
            } finally {
                try {
                    jbiInputStream.close();
                } catch (final IOException e) {
                    throw new DeploymentException(e);
                }
            }
        } catch (final CDKJBIDescriptorException e) {
            throw new DeploymentException(e);
        } catch (FileNotFoundException e) {
            throw new DeploymentException(e);
        }

        ServiceUnitDataHandler handler = null;
        try {
            if (this.component.getComponentConfiguration().getPropertiesFile() != null
                    && !StringHelper.isNullOrEmpty(this.component.getComponentConfiguration()
                            .getPropertiesFile().getValue())) {
                handler = new ServiceUnitDataHandler(serviceUnitName, suRootPath, jbiDescriptor,
                        this.component.getComponentConfiguration().getPropertiesFile().getValue());
            }
        } catch (PEtALSCDKException e1) {
            this.logger
                    .warning("Failed to initialize the SU parameters against the component properties file at '"
                            + this.component.getComponentConfiguration().getPropertiesFile()
                            + "'. Ignore the properties.");
        }

        if (handler == null) {
            handler = new ServiceUnitDataHandler(serviceUnitName, suRootPath, jbiDescriptor);
        }

        // validate the SU extensions if validator is available.
        // TODO : Deprecated when the component specific XSD validation
        // would be available
        if (this.component.getServiceUnitExtensionsValidator() != null) {
            try {
                this.validateExtensions(jbiDescriptor);
            } catch (final PEtALSCDKException e) {
                throw new DeploymentException(
                        "Failed to validate at least one SU extensions for SU " + serviceUnitName,
                        e);
            }
        }

        try {
            this.processConsumesNodes(handler);
        } catch (final JBIException ex) {
            throw new DeploymentException("Failed to process consumes nodes for SU "
                    + serviceUnitName, ex);
        }

        // Now we can register the service unit
        synchronized (this.serviceUnitDataHandlers) {
            if (this.serviceUnitDataHandlers.containsKey(serviceUnitName)) {
                throw new DeploymentException(
                        "A service unit with the given name is already registered : "
                                + serviceUnitName);
            } else {
                this.logger
                        .log(Level.FINEST, "Register a data handler for SU : " + serviceUnitName);
                this.serviceUnitDataHandlers.put(serviceUnitName, handler);

                // additional deployment by component implementation
                if (this.component.getServiceUnitListener() != null) {
                    try {
                        this.component.getServiceUnitListener().onSUDeployed(serviceUnitName,
                                suRootPath, jbiDescriptor);
                    } catch (final Throwable e) {
                        // clean cache if any exception
                        this.logger.log(Level.FINEST, "Unregister a data handler for SU : "
                                + serviceUnitName);
                        this.serviceUnitDataHandlers.remove(serviceUnitName);

                        throw new DeploymentException(
                                "The ServiceUnit listener failed to process for SU "
                                        + serviceUnitName, e);
                    }
                }
            }
        }

        final String result = ManagementMessageUtil.getComponentTaskResult(this.component
                .getContext().getComponentName(), "deploy",
                ManagementMessageUtil.TASK_RESULT_SUCCESS);
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     */
    public void start(final String serviceUnitName) throws DeploymentException {
        if (this.component.getContext() == null) {
            // the component has not been initialized for the first time...
            throw new DeploymentException(
                    "Can not start a service unit on a component which has not been initialized");
        }

        this.logger.log(Level.FINE, "start serviceUnitName  " + serviceUnitName);

        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers.get(serviceUnitName);
        try {
            this.processProvidesNodes(handler);
        } catch (final JBIException ex) {
            final String msg = "Failed to start SU " + serviceUnitName;
            this.logger.log(Level.SEVERE, msg, ex);
            throw new DeploymentException(msg, ex);
        }

        // call the Service Unit Listener if any
        if (this.component.getServiceUnitListener() != null) {
            try {
                this.component.getServiceUnitListener().onSUStarted(serviceUnitName);
            } catch (final PEtALSCDKException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process for SU "
                        + serviceUnitName, e);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     */
    public void stop(final String serviceUnitName) throws DeploymentException {
        if (this.component.getContext() == null) {
            // the component has not been initialized for the first time...
            throw new DeploymentException(
                    "Can not stop a service unit on a component which has not been initialized");
        }

        this.logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);

        if (this.component.getServiceUnitListener() != null) {
            try {
                this.component.getServiceUnitListener().onSUStopped(serviceUnitName);
            } catch (final PEtALSCDKException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process for SU "
                        + serviceUnitName, e);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#shutDown(java.lang.String)
     */
    public void shutDown(final String serviceUnitName) throws DeploymentException {
        // stop the providing of services for the given SU, as said in the JBI
        // specifications
        if (this.component.getContext() == null) {
            // the component has not been initialized for the first time...
            throw new DeploymentException(
                    "Can not shutdown a service unit on a component which has not been initialized");
        }

        this.logger.log(Level.FINE, "shutdown  serviceUnitName  " + serviceUnitName);

        try {
            // No synchronization on serviceUnitDataHandlers is needed because
            // it's a ConcurrentMap
            final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers
                    .get(serviceUnitName);
            final Iterator<ServiceEndpoint> endpoints = handler.getServiceEndpoints().values()
                    .iterator();
            while (endpoints.hasNext()) {
                final ServiceEndpoint endpoint = endpoints.next();

                this.component.getContext().deactivateEndpoint(endpoint);

                // The endpoint removing must be done on the Iterator, not on
                // the Map.
                // Otherwise a ConcurrentModificationException will occurs when
                // shutdowning a SA containing a SU containing more than one
                // service.
                endpoints.remove();
            }

        } catch (final JBIException e) {
            throw new DeploymentException("Deactivate endpoint failed for SU " + serviceUnitName, e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String,
     * java.lang.String)
     */
    public String undeploy(final String serviceUnitName, final String serviceUnitRootPath)
            throws DeploymentException {
        if (this.component.getContext() == null) {
            // the component has not been initialized for the first time...
            throw new DeploymentException(
                    "Can not undeploy a service unit on a component which has not been initialized");
        }

        synchronized (this.serviceUnitDataHandlers) {
            final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers
                    .get(serviceUnitName);
            if (handler != null) {
                try {
                    if (this.component.getServiceUnitListener() != null) {
                        this.component.getServiceUnitListener().onSUUndeployed(serviceUnitName,
                                serviceUnitRootPath, handler.getDescriptor());
                    }
                } catch (final PEtALSCDKException e) {
                    // instead of throwing an exception, return an error message
                    return ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                            .getComponentName(), "undeploy",
                            ManagementMessageUtil.TASK_RESULT_FAILED,
                            ManagementMessageUtil.MESSAGE_TYPE_ERROR, "ERROR",
                            "The ServiceUnit manager failed to process", new String[] {
                                    serviceUnitName, serviceUnitRootPath }, e.getMessage());
                } finally {
                    this.logger.log(Level.FINEST, "Unregister a data handler for SU "
                            + serviceUnitName);
                    this.serviceUnitDataHandlers.remove(serviceUnitName);
                }
        
                return ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                        .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
            } else {
                return ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                        .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_FAILED,
                        ManagementMessageUtil.MESSAGE_TYPE_ERROR, "ERROR",
                        "The ServiceUnit manager failed to process", new String[] { serviceUnitName,
                                serviceUnitRootPath }, "Unknown service unit");
            }
        }
    }

    /**
     * Return all {@link ServiceUnitDataHandler}s as a Map
     * 
     * @return all {@link ServiceUnitDataHandler}s as a Map
     */
    public Map<String, ServiceUnitDataHandler> getServiceUnitDataHandlers() {
        return this.serviceUnitDataHandlers;
    }

    /**
     * Retrieve the Provides node that has activated the given jbi Endpoint
     * 
     * @param serviceEndpoint
     *      the previously activated jbi Endpoint
     * @return the {@link Provides} element
     */
    public Provides getProvidesFromEndpoint(final ServiceEndpoint serviceEndpoint) {
        if (serviceEndpoint != null) {
            final ServiceEndpointKey serviceEndpointKey = new ServiceEndpointKey(serviceEndpoint
                    .getServiceName(), serviceEndpoint.getEndpointName());
            // No synchronization on serviceUnitDataHandlers is needed because
            // it's a ConcurrentMap
            for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
                final Provides provides = dataHandler.getEpJBIDesc().get(serviceEndpointKey);
                if (provides != null) {
                    return provides;
                }
            }
        }
        return null;
    }

    /**
     * Retrieve the Consumes node matching the given destination information
     * (endpoint name, service name, interface name). Search order is by
     * ascending priority : endpointName, serviceName, interfaceName. If there
     * are many consumes matching information, return the first one found.
     * 
     * @param endpointName
     *      an endpoint name, can be null
     * @param serviceName
     *      a service {@link QName}, can be null
     * @param interfaceName
     *      an interface {@link QName}, can be null
     * @return the {@link Consumes} element matching information, can be null
     */
    public Consumes getConsumesFromDestination(final String endpointName, final QName serviceName,
            final QName interfaceName) {
        Consumes result = null;
        if (endpointName != null) {
            result = this.getConsumesForEndpoint(endpointName);
        }
        if (result == null) {
            if (serviceName != null) {
                result = this.getConsumesForService(serviceName);
            }
            if (result == null) {
                if (interfaceName != null) {
                    result = this.getConsumesForInterface(interfaceName);
                }
            }
        }
        return result;
    }

    /**
     * Retrieve the Service description (as DOM Document) of the given jbi
     * Endpoint
     * 
     * @param serviceEndpoint
     *      the jbi Endpoint
     * @return the service description as Dom Document
     */
    public Document getServiceDescription(final ServiceEndpoint serviceEndpoint) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            Document serviceDescription = dataHandler.getEndpointDescription(serviceEndpoint);
            if (serviceDescription != null) {
                return serviceDescription;
            }
        }
        return null;
    }

    /**
     * Get the SU Data Handler matching the given Consumes/Provides service.
     * 
     * @param consumes
     *      The @ code Consumes} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandlerForService(final Object service) {
        ServiceUnitDataHandler suDataHandler = null;

        if (service instanceof Consumes) {
            suDataHandler = this.getSUDataHandlerForConsumes((Consumes) service);
        } else if (service instanceof Provides) {
            suDataHandler = this.getSUDataHandlerForProvides((Provides) service);
        }

        return suDataHandler;
    }

    /**
     * Get the SU Data Handler matching the given Provides.
     * 
     * @param consumes
     *      The @ code Consumes} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandlerForConsumes(final Consumes consumes) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (Consumes consumesElement : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (consumes.equals(consumesElement)) {
                    return dataHandler;
                }
            }
        }
        return null;
    }

    /**
     * Get the SU Data Handler matching the given Consumes.
     * 
     * @param provides
     *      The @ code Provides} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandlerForProvides(final Provides provides) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (Provides providesElement : dataHandler.getDescriptor().getServices().getProvides()) {
                if (provides.equals(providesElement)) {
                    return dataHandler;
                }
            }
        }
        return null;
    }

    /**
     * Search the wsdl description of the activated endpoint in the Provides
     * extensions. If no wsdl description can be found, it creates a light
     * weight Wsdl description from Provides Interface, Service and Endpoint
     * names. No generation of endpoint is processed to know after the generated
     * name to update the provides too.
     * 
     * 
     * @param wsdlLocation
     *            the location of the wsdl document. Can be {@code null}
     * @param provides
     *            the Provides node
     * @return the WSDL description as DOM Document
     * @throws PEtALSCDKException
     */
    protected Document getWsdlDescription(final URL wsdlLocation, final Provides provides)
            throws PEtALSCDKException {
        
        final Settableboolean wsdlImportsDownload = provides.getWsdlImportsDownload();
        boolean isWsdlImportsDownload = wsdlImportsDownload == null ? true : wsdlImportsDownload
                .isValue();
        
        if (wsdlLocation != null) {
            return WSDLHelper.createDocumentFromWSDL(wsdlLocation,
                    isWsdlImportsDownload);
        } else {
            // if the WSDL field is null, generate a basic WSDL
            return WSDLHelper.createLightWSDL20(provides.getInterfaceName(), provides
                    .getServiceName(), provides.getEndpointName());
        }
    }

    /**
     * Get the location of the wsdl file.
     * 
     * @param installRoot
     *            the root install used to search locally the file if needed
     * @param provides
     *            the provides read to get the wsdl file name
     * @return the valid URL of the Wsdl file, or null
     * @throws PEtALSCDKException
     */
    protected URL getWsdlLocation(final String installRoot, final Provides provides)
            throws PEtALSCDKException {
        
        URL wsdlLocation = null;
        final String declaredWsdlPath = provides.getWsdl();
        if (StringHelper.isNullOrEmpty(declaredWsdlPath)) {
            wsdlLocation = null;
        }
        else {
            try {
                wsdlLocation = new URL(declaredWsdlPath);
            } catch (final MalformedURLException e) {
                // Can be a local file
                final File wsdlFile = new File(installRoot, declaredWsdlPath);
                if (!wsdlFile.exists()) {
                    // Try using the classloader
                    wsdlLocation = Thread.currentThread().getContextClassLoader().getResource(declaredWsdlPath);
                    if (wsdlLocation == null) {
                        throw new PEtALSCDKException("Can not retrieve WSDL document from location : "
                            + declaredWsdlPath);
                    }
                }
                else {
                    try {
                        wsdlLocation = wsdlFile.toURI().toURL();
                    } catch (final MalformedURLException e1) {
                        // not possible
                    }
                }
            }
        }
        return wsdlLocation;
    }
    
    /**
     * Process the interceptors for a consumes/provides service.
     * 
     * @param handler
     * @param service
     *      the {@code Provides} or {@code Consumes} service
     * @throws JBIException
     */
    protected void processInterceptors(final ServiceUnitDataHandler handler, final Object service)
            throws JBIException {

        SUInterceptors suInterceptors = null;

        if (service instanceof Consumes) {
            suInterceptors = ((Consumes) service).getSuInterceptors();
        } else if (service instanceof Provides) {
            suInterceptors = ((Provides) service).getSuInterceptors();
        }

        // create the manager for this extension element. Each element has a
        // linked manager with preloaded interceptors.

        final MessageInterceptorManager interceptorManager;
        if (this.component.getComponentConfiguration().getPropertiesFile() != null
                && !StringHelper.isNullOrEmpty(this.component.getComponentConfiguration()
                        .getPropertiesFile().getValue())) {
            interceptorManager = new MessageInterceptorManager(this.component
                    .getComponentConfiguration().getPropertiesFile().getValue(), this.logger);
        } else {
            interceptorManager = new MessageInterceptorManager(this.logger);
        }

        List<SUInterceptor> sendSUInterceptors = null;
        List<SUInterceptor> sendResponseSUInterceptors = null;
        List<SUInterceptor> acceptSUInterceptors = null;
        List<SUInterceptor> acceptResponseSUInterceptors = null;
        if (suInterceptors != null) {
            if (suInterceptors.getSend() != null) {
                sendSUInterceptors = suInterceptors.getSend().getInterceptor();
            }
            if (suInterceptors.getSendResponse() != null) {
                sendResponseSUInterceptors = suInterceptors.getSendResponse().getInterceptor();
            }
            if (suInterceptors.getAccept() != null) {
                acceptSUInterceptors = suInterceptors.getAccept().getInterceptor();
            }
            if (suInterceptors.getAcceptResponse() != null) {
                acceptResponseSUInterceptors = suInterceptors.getAcceptResponse().getInterceptor();
            }
        }

        interceptorManager.init(this.component.getInterceptors(), acceptSUInterceptors,
                sendSUInterceptors, acceptResponseSUInterceptors, sendResponseSUInterceptors);

        handler.addMessageInterceptorManager(service, interceptorManager);

    }

    /**
     * 
     * @param handler
     * @throws JBIException
     */
    protected void processConsumesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        final Jbi descriptor = handler.getDescriptor();
        final List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            this.processInterceptors(handler, consumes);
        }
    }

    /**
     * Process the given provides node.
     * 
     * @param handler
     * @param provides
     * @return The resulted Service Endpoint
     * @throws JBIException
     */
    protected ServiceEndpoint processProvidesNode(final ServiceUnitDataHandler handler,
            final Provides provides) throws JBIException {

        final URL wsdlLocation = this.getWsdlLocation(handler.getInstallRoot(),
                provides);
        final Document wsdlDescription = this.getWsdlDescription(wsdlLocation, provides);
        
        String endpointName = provides.getEndpointName();
        
        if (AUTOGENERATED_ENDPOINT_NAME.equals(endpointName)) {
            endpointName = IDGenerator.getInstance().getNewID();

            // update the provides object to set the real endpoint
            provides.setEndpointName(endpointName);

            final File jbiXmlFile = new File(handler.getInstallRoot() + File.separator + META_INF
                    + File.separator + JBI_XML);
            
            // If the wsdl is loaded from a Jar file we must externalized the
            // WSDL
            URL locationToSaveWSDL;
            if ("jar".equals(wsdlLocation.getProtocol())) {
                // WSDL loaded from a JAR File
                final String filename = new File(wsdlLocation.getFile()).getName();
                try {
                    locationToSaveWSDL = new File(handler.getInstallRoot(), filename).toURI()
                            .toURL();
                    provides.setWsdl(filename);
                } catch (final MalformedURLException e1) {
                    this.logger.log(Level.SEVERE, "The following error has not to occur", e1);
                    locationToSaveWSDL = wsdlLocation;
                }
            } else {
                locationToSaveWSDL = wsdlLocation;
            }
            

            // update the descriptor of the SU to set the real endpoint
            // then, if the SU is restarted, the real endpoint is
            // used
            try {
                JBIDescriptorBuilder.buildXmlJBIdescriptor(handler.getDescriptor(),
                        new FileOutputStream(jbiXmlFile));
            } catch (FileNotFoundException e) {
                throw new JBIException("Failed to persist the auto-generated endpoint name", e);
            } catch (CDKJBIDescriptorException e) {
                throw new JBIException("Failed to persist the auto-generated endpoint name", e);
            }
            
            // Update WSDL to set the endpoint name
            this.updateWSDLWithEndpointName(wsdlDescription, provides.getServiceName(),
                    endpointName);
            
            // Save the updated WSDL
            if (wsdlLocation != null) {
                
                try {
                    final File wsdlSavedFile = new File(locationToSaveWSDL.getFile());
                    final OutputStream wsdlOutputStream = new FileOutputStream(wsdlSavedFile);
                    try {
                        WSDLHelper.saveWSDLImports(wsdlDescription, wsdlSavedFile.getParentFile());
                        WSDLHelper.printWSDL(WSDLHelper.removeImportedDocuments(wsdlDescription), wsdlOutputStream);
                        
                    } finally {
                        wsdlOutputStream.close();
                    }
                } catch (final FileNotFoundException e) {
                    this.logger.severe("Unable to persist updated service description (WSDL) '"
                            + locationToSaveWSDL + "' (" + e.getMessage() + ").");
                } catch (final IOException e) {
                    this.logger.severe("Unable to persist updated service description (WSDL) '"
                            + locationToSaveWSDL + "' (" + e.getMessage() + ").");
                }
                
            }
        }

        handler.addEndpointDescription(provides.getServiceName(), endpointName, wsdlDescription);
        handler.addJBIDescription(provides.getServiceName(), endpointName, provides);
        final ServiceEndpoint ep = this.component.getContext().activateEndpoint(
                provides.getServiceName(), endpointName);
        handler.addServiceEndpoint(ep);
        
        this.processInterceptors(handler, provides);

        return ep;
    }
    
    private void updateWSDLWithEndpointName(final Document wsdl, final QName service,
            final String newEndpointName) {
        final Element rootElement = wsdl.getDocumentElement();

        // Retrieve the service
        final NodeList services = rootElement.getElementsByTagNameNS(
                "http://schemas.xmlsoap.org/wsdl/", "service");

        if (services.getLength() > 0) {
            for (int i = 0; i < services.getLength(); i++) {
                final Element serviceElement = (Element) services.item(i);
                if (serviceElement.getAttribute("name").equals(service.getLocalPart())) {
                    // Retrieve the endpoint name to update the first one.
                    final NodeList endpoints = serviceElement.getElementsByTagNameNS(
                            "http://schemas.xmlsoap.org/wsdl/", "port");
                    if (endpoints.getLength() > 0) {
                        ((Element) endpoints.item(0)).setAttribute("name", newEndpointName);
                    }
                    break;
                }
            }
        }

    }

    /**
     * 
     * @param handler
     * @throws JBIException
     */
    protected void processProvidesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        final Jbi descriptor = handler.getDescriptor();

        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            ServiceEndpoint sep = this.processProvidesNode(handler, provides);
            this.logger.log(Level.INFO, "New Service Endpoint has been deployed : " + sep);
        }
    }

    /**
     * Validate the SU extensions.
     * 
     * @param descriptor
     *      the service unit descriptor
     * @throws PEtALSCDKException
     *      if at least an extension block is not valid. It is the role of the
     *      developer to specify which extension is not valid.
     */
    protected void validateExtensions(final Jbi descriptor) throws PEtALSCDKException {

        // validate the provides
        // TODO : Deprecated when the component specific XSD validation would be
        // available
        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            this.component.getServiceUnitExtensionsValidator().validateProvidesExtensions(
                    new ConfigurationExtensions(provides.getAny()));
        }

        // validate the consumes
        // TODO : Deprecated when the component specific XSD validation would be
        // available
        final List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            this.component.getServiceUnitExtensionsValidator().validateConsumesExtensions(
                    new ConfigurationExtensions(consumes.getAny()));
        }
    }

    /**
     * Retrieve the install root of the Service Unit that has activated the
     * given endpoint
     * 
     * @param endpoint
     *      The activated endpoint
     * @return the Service unit install root
     */
    protected String getSUInstallRootForActivatedEp(final ServiceEndpoint endpoint) {
        return this.getSUDataHandlerForEndpoint(endpoint).getInstallRoot();
    }

    /**
     * Retrieve the data handler of the Service Unit that has activated the
     * given endpoint.
     * 
     * @param endpoint
     *      the
     * @return the {@link ServiceUnitDataHandler}
     */
    public ServiceUnitDataHandler getSUDataHandlerForEndpoint(final ServiceEndpoint endpoint) {
        final ServiceEndpointKey key = new ServiceEndpointKey(endpoint.getServiceName(), endpoint
                .getEndpointName());
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final ServiceEndpointKey k : dataHandler.getEpJBIDesc().keySet()) {
                if (key.equals(k)) {
                    return dataHandler;
                }
            }
        }

        return null;
    }

    /**
     * Get the consumes matching the given endpoint name.
     * 
     * @param serviceName
     *      The endpoint name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    public Consumes getConsumesForEndpoint(final String endpointName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (ServiceUnitDataHandler dataHandler : serviceUnitDataHandlers.values()) {
            for (Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (endpointName.equals(consumes.getEndpointName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the consumes matching the given service name.
     * 
     * @param serviceName
     *      The service qualified name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    public Consumes getConsumesForService(final QName serviceName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (ServiceUnitDataHandler dataHandler : serviceUnitDataHandlers.values()) {
            for (Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (serviceName.equals(consumes.getServiceName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the consumes matching the given service name.
     * 
     * @param serviceName
     *      The service qualified name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    public Consumes getConsumesForInterface(final QName interfaceName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (ServiceUnitDataHandler dataHandler : serviceUnitDataHandlers.values()) {
            for (Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (interfaceName.equals(consumes.getInterfaceName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the {@code set} of endpoints exposed by the given Service Unit.
     * 
     * @param serviceUnitName
     *      The name of the Service Unit
     * @return a {@code set} of {@code ServiceEndpoint}
     */
    public Set<ServiceEndpoint> getEndpointsForServiceUnit(final String serviceUnitName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        final Set<ServiceEndpoint> result = new HashSet<ServiceEndpoint>();
        for (final ServiceEndpointKey key : this.serviceUnitDataHandlers.get(serviceUnitName)
                .getServiceEndpoints().keySet()) {
            result.add(this.serviceUnitDataHandlers.get(serviceUnitName).getServiceEndpoints().get(
                    key));
        }
        return result;
    }
}
